from collections import namedtuple
from django.db.models.query import QuerySet
from typing import List


class CustomFilterWizardInterface:
    """
        Interface responsible for delivering signatures of methods for
        building custom filters for a particular kind of object.

        Constants:
            FILTERS_BASED_ON_FIELD_TYPE: kind of mapper to indicate what kind of filters should be used
            depends on the type of the particular field.
    """

    FILTERS_BASED_ON_FIELD_TYPE = {
        "string": ["iexact", "istartswith", "icontains"],
        "integer": ["exact", "lt", "lte", "gt", "gte"],
        "decimal": ["exact", "lt", "lte", "gt", "gte"],
        "date": ["exact", "lt", "lte", "gt", "gte"],
        "boolean": ["exact"],
    }

    def get_type_of_object(self) -> str:
        """
        Get the type of object for which we want to define a specific way of building filters.

        :return: The type of the object.
        :rtype: str
        """
        pass

    def load_definition(self, tuple_type: type, **kwargs) -> List[namedtuple]:
        """
        Load the definition of how to create filters.

        This method retrieves the definition of how to create filters and returns it as a list of named tuples.
        Each named tuple is built with the provided `tuple_type` and has the fields `field`, `filter`, and `value`.

        Example named tuple: <Type>(field=<str>, filter=<str>, value=<str>)
        Example usage: BenefitPlan(field='income', filter='lt, gte, icontains, exact', value='')

        :param tuple_type: The type of the named tuple.
        :type tuple_type: type

        :return: A list of named tuples representing the definition of how to create filters.
        :rtype: List[namedtuple]
        """
        pass

    def apply_filter_to_queryset(self, custom_filters: List[namedtuple], query: QuerySet, relation=None):
        """
        Apply custom filters to a queryset.

        :param custom_filters: List of named tuples representing custom filters.
        :type custom_filters: List[namedtuple]

        :param query: The original queryset with filters.
        :type query: django.db.models.query.QuerySet

        :param relation: The optional argument which defines the relation field in queryset for example 'beneficiary'
        :type relation: str or None

        :return: The updated queryset with additional filters applied.
        :rtype: django.db.models.query.QuerySet
        """
        pass



import functools
from inspect import getfullargspec


class ObligatoryFieldValidationError(Exception):
    ...


class ObligatoryFieldValidation:
    def __init__(self, obligatory_fields_list):
        self.obligatory_field_list = obligatory_fields_list

    def validate_obligatory_fields(self, payload):
        """
        Validate payload against obligatory field definitions. Payload have to share same fields as model in which
        data is validated. Validation is done using obligatory_fields_list. If field from obligatory field list is
        not provided as key in payload or is blank then ObligatoryFieldValidationError exception will be raised.

        :param payload: dictionary from which object will be created.
        :return: None
        """
        for field_name, control in self.obligatory_field_list.items():
            if control not in ('O', 'H', 'M'):
                raise ObligatoryFieldValidationError(
                    F'Invalid configuration for field {field_name}, value {control}'
                    F'is not proper value. Allowed values are:'
                    F'- [O] - Optional\n'
                    F'- [H] - Hidden\n'
                    F'- [M] - Mandatory\n'
                    F'Contact Administrator to fix this.'
                )
            if control == 'O':
                return
            elif control == 'H':
                if payload.get(field_name):
                    raise ObligatoryFieldValidationError(
                        F'Field {field_name} is set as hidden, but payload provides value. '
                        F'Does field {field_name} have default?')
            elif control == 'M' and not (payload.get(field_name)):
                raise ObligatoryFieldValidationError(F'Field {field_name} is mandatory, '
                                                     F'but payload does not provide value')


def validate_payload_for_obligatory_fields(list_of_obligatory_fields, payload_arg='data'):
    """
    Function decorator used for conveniently validate payload in functions.
    It creates instance of ObligatoryFieldValidation using list_of_obligatory_fields argument as obligatory
    characters and executes validation.
    :param list_of_obligatory_fields: List of obligatory fields for designated payload.
    :param payload_arg: Additional argument determining name of decorated function argument that provides payload.
    :return: None
    """

    def decorator(func):
        # Get index of payload arg in case it's not explicitly provided as kwarg.
        argspec = getfullargspec(func)
        argument_index = argspec.args.index(payload_arg)

        @functools.wraps(func)
        def wrapper_validate_fields(*args, **kwargs):
            try:
                value = args[argument_index]
            except IndexError:
                # Payload passed as kwarg not arg
                value = kwargs[payload_arg]

            validator = ObligatoryFieldValidation(list_of_obligatory_fields)
            validator.validate_obligatory_fields(value)
            out = func(*args, **kwargs)
            return out

        return wrapper_validate_fields

    return decorator



# Generated by Django 3.2.24 on 2024-04-09 10:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0027_alter_interactiveuser_last_login_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='moduleconfiguration',
            name='module',
            field=models.CharField(max_length=80),
        ),
    ]



from calendar import timegm

import jwt
from graphql_jwt.settings import jwt_settings
from graphql_jwt.signals import token_issued
from django.apps import apps
from django.utils import timezone
from django.dispatch import receiver
import logging
import uuid
from datetime import datetime

logger = logging.getLogger(__file__)


@receiver(token_issued)
def on_token_issued(sender, request, user, **kwargs):
    # Store the date on which the user got the auth token
    user.last_login = timezone.now()
    user.save()
    pass


def jwt_encode_user_key(payload, context=None):
    payload['jti'] = str(uuid.uuid4())
    payload['nbf'] = timegm(datetime.utcnow().utctimetuple())

    token = jwt.encode(
        payload,
        get_jwt_key(encode=True, context=context, payload=payload),
        algorithm=jwt_settings.JWT_ALGORITHM,
    )
    # JWT module after 1.7 does the encoding, introducing some conflicts in graphql-jwt, let's support both
    if isinstance(token, bytes):
        token = token.decode("utf-8")
    return token


def jwt_decode_user_key(token, context=None):
    # First decode the token without validating it, so we can extract the username
    not_validated = jwt.decode(
        token,
        get_jwt_key(encode=False, context=context),
        options={
            'verify_exp': jwt_settings.JWT_VERIFY_EXPIRATION,
            'verify_aud': jwt_settings.JWT_AUDIENCE is not None,
            'verify_signature': False,
        },
        leeway=jwt_settings.JWT_LEEWAY,
        audience=jwt_settings.JWT_AUDIENCE,
        issuer=jwt_settings.JWT_ISSUER,
        algorithms=[jwt_settings.JWT_ALGORITHM],
    )
    if not_validated and not_validated.get("username"):
        user_class = apps.get_model("core", "User")
        db_user = user_class.objects \
            .filter(username=not_validated.get("username")) \
            .only("i_user__private_key") \
            .first()
        if db_user and db_user.private_key:
            key = db_user.private_key
        else:
            key = get_jwt_key(encode=False)
    else:
        key = get_jwt_key(encode=False)
    return jwt.decode(
        token,
        key,
        options={
            'verify_exp': jwt_settings.JWT_VERIFY_EXPIRATION,
            'verify_aud': jwt_settings.JWT_AUDIENCE is not None,
            'verify_signature': jwt_settings.JWT_VERIFY,
        },
        leeway=jwt_settings.JWT_LEEWAY,
        audience=jwt_settings.JWT_AUDIENCE,
        issuer=jwt_settings.JWT_ISSUER,
        algorithms=[jwt_settings.JWT_ALGORITHM],
    )


def get_jwt_key(encode=True, context=None, payload=None):
    user_key = extract_private_key_from_context(context)
    if user_key is None and payload is not None:
        user_key = extract_private_key_from_payload(payload)
    if user_key:
        return user_key

    if encode:
        return getattr(jwt_settings, "JWT_PRIVATE_KEY", jwt_settings.JWT_SECRET_KEY)
    else:
        return getattr(jwt_settings, "JWT_PUBLIC_KEY", jwt_settings.JWT_SECRET_KEY)


def extract_private_key_from_payload(payload):
    # Get user private key from payload. This covers the refresh token mutation
    from core.models import User

    if "username" in payload:
        return User.objects.get(username=payload["username"]).private_key


def extract_private_key_from_context(context):
    if context and context.user and hasattr(context.user, "private_key"):
        return context.user.private_key
    return None



from django.dispatch import receiver
import sys
from django.apps import apps
from django.db.models.signals import post_save, post_delete
from contextlib import suppress
from core.models.user import Officer, Role
from django.core.cache import cache
from django_redis.cache import RedisCache


@receiver([post_save, post_delete], sender=Officer)
def _post_save_eo_receiver(sender, instance, **kwargs):
    with suppress(AttributeError):
        cache.delete(f"user_eo_{instance.code}")
        cache.delete(f"rights_{instance.code}")

@receiver([post_save, post_delete], sender=Role)
def _post_save_rolerights_receiver(sender, instance, **kwargs):
    with suppress(AttributeError):
        if isinstance(cache, RedisCache):
            cache.delete("rights_*")
        else:
            cache.clear()
        
if 'claim' in sys.modules:
    ClaimAdmin = apps.get_model('core', 'ClaimAdmin')
    @receiver([post_save, post_delete], sender=ClaimAdmin)
    def _post_save_ca_receiver(sender, instance, **kwargs):
        with suppress(AttributeError):
            cache.delete(f"user_ca_{instance.code}")
            cache.delete(f"rights_{instance.code}")



from django.contrib.auth.middleware import (
    RemoteUserMiddleware as dj_RemoteUserMiddleware,
)
from rest_framework import permissions
from os import environ
from rest_framework.permissions import IsAuthenticated


def checkUserWithRights(rights):
    class UserWithRights(IsAuthenticated):
        def has_permission(self, request, view):
            return super().has_permission(request, view) and request.user.has_perms(
                rights
            )

    return UserWithRights


class ObjectPermissions(permissions.DjangoObjectPermissions):
    perms_map = {
        "GET": ["%(app_label)s.view_%(model_name)s"],
        "OPTIONS": ["%(app_label)s.view_%(model_name)s"],
        "HEAD": ["%(app_label)s.view_%(model_name)s"],
        "POST": ["%(app_label)s.add_%(model_name)s"],
        "PUT": ["%(app_label)s.change_%(model_name)s"],
        "PATCH": ["%(app_label)s.change_%(model_name)s"],
        "DELETE": ["%(app_label)s.delete_%(model_name)s"],
    }


class RemoteUserMiddleware(dj_RemoteUserMiddleware):
    header = environ.get("REMOTE_USER_HEADER_NAME", "HTTP_REMOTE_USER")



import csv

from django.http import Http404, StreamingHttpResponse
from django.views.decorators.http import require_GET
from isodate import strftime
from rest_framework import viewsets, status
from rest_framework.decorators import action, api_view
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from .models import User, ExportableQueryModel
from .scheduler import scheduler
from .serializers import UserSerializer
from django.utils.translation import gettext as _


def check_user_rights(rights):
    class UserWithRights(IsAuthenticated):
        def has_permission(self, request, view):
            return super().has_permission(request, view) and request.user.has_perms(rights)

    return UserWithRights

class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer
    # If we don't specify the IsAuthenticated, the framework will look for the core.user_view permission and prevent
    # any access from non-admin users
    permission_classes = [IsAuthenticated]

    @action(detail=False)
    def current_user(self, request):
        serializer = self.get_serializer(request.user, many=False)
        return Response(serializer.data)


@api_view(['GET'])
@require_GET
def fetch_export(request):
    requested_export = request.query_params.get('export')
    export = ExportableQueryModel.objects.filter(name=requested_export).first()
    if not export:
        raise Http404
    elif export.user != request.user:
        raise PermissionDenied({"message": _("Only user requesting export can fetch request")})
    elif export.is_deleted:
        return Response(data='Export csv file was removed from server.', status=status.HTTP_410_GONE)

    export_file_name = F"export_{export.model}_{strftime(export.create_date, '%d_%m_%Y')}.{export.file_format}"
    if export.file_format == ExportableQueryModel.FileFormat.CSV:
        response = StreamingHttpResponse(
            open(export.content.path, 'rb'),
            content_type="text/csv",
            headers={'Content-Disposition': F'attachment; filename="{export_file_name}"'},
        )
    elif ExportableQueryModel.FileFormat.XLSX:
        response = StreamingHttpResponse(
            open(export.content.path, 'rb'),
            content_type="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            headers={'Content-Disposition': F'attachment; filename="{export_file_name}"'},
        )
    else:
        return Response(data='Unsupported file format.', status=status.HTTP_400_BAD_REQUEST)

    return response

def _serialize_job(job):
    return "name: %s, trigger: %s, next run: %s, handler: %s" % (
        job.name, job.trigger, job.next_run_time, job.func)


@api_view(['GET'])
@require_GET
def get_scheduled_jobs(request):
    return Response([_serialize_job(job) for job in scheduler.get_jobs()])



from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _


class ObjectExistsValidationMixin:
    INVALID_UPDATE_ID_MSG = _("%(model)s for id  %(id)s does not exists")

    @classmethod
    def validate_object_exists(cls, id_):
        existing = cls.OBJECT_TYPE.objects.filter(id=id_).first()
        if not existing:
            raise ValidationError(cls.INVALID_UPDATE_ID_MSG % {'id': id_, 'model': str(cls.OBJECT_TYPE)})



from functools import wraps

from core.apps import CoreConfig
from core.data_masking import MaskingClassStorage

def anonymize_gql():
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            queryset = result.iterable
            model = queryset.model
            edges = result.edges
            user = kwargs.get('user')
            if not user.has_perms(CoreConfig.gql_query_enable_viewing_masked_data_perms):
                if model:
                    masking_class = MaskingClassStorage.get_masking_class(model.__name__)
                    if masking_class:
                        masking_enabled = masking_class.masking_enabled
                        if masking_enabled:
                            for edge in edges:
                                masking_class.apply_mask(edge.node)
            return result
        return wrapper
    return decorator
